import sys
import os

from hokusai.cli import *
from hokusai.lib.environment import cert_file_path

path = cert_file_path()
if path is not None:
  os.environ['SSL_CERT_FILE'] = path

def main():
  base(obj={})



import copy
import json
import os
import pipes
import re

from hokusai import CWD
from hokusai.lib.common import (
  clean_dict,
  file_debug,
  k8s_uuid,
  key_value_list_to_dict,
  returncode,
  shout,
  user,
  validate_key_value
)
from hokusai.lib.config import (
  config,
  HOKUSAI_CONFIG_DIR
)
from hokusai.lib.exceptions import HokusaiError
from hokusai.lib.template_selector import TemplateSelector
from hokusai.services.ecr import ECR
from hokusai.services.kubectl import Kubectl
from hokusai.services.yaml_spec import YamlSpec


class CommandRunner:
  def __init__(self, context, namespace=None):
    self.context = context
    self.kctl = Kubectl(self.context, namespace=namespace)
    self.ecr = ECR()
    self.pod_name = self._name()
    self.container_name = self.pod_name
    self.yaml_template = TemplateSelector().get(
      os.path.join(CWD, HOKUSAI_CONFIG_DIR, context)
    )
    self.model_deployment = config.run_template

  def _clean_containers_spec(self, spec):
    '''
    given a pod's 'containers' spec, return that spec,
    but ensure that each container spec has only the desired fields.
    '''
    container_fields_to_keep = [
      'args',
      'envFrom',
      'image',
      'imagePullPolicy',
      'name',
      'volumeMounts'
    ]
    clean_spec = []
    for container_spec in spec:
      clean_spec += [
        clean_dict(container_spec, container_fields_to_keep)
      ]
    return clean_spec

  def _clean_pod_spec(self, spec):
    '''
    given a pod spec, return that spec,
    but ensure that it has only the desired fields.
    '''
    pod_fields_to_keep = [
      'containers',
      'dnsConfig',
      'dnsPolicy',
      'initContainers',
      'serviceAccountName',
      'volumes'
    ]
    clean_spec = clean_dict(
      spec, pod_fields_to_keep
    )
    file_debug(
      clean_spec,
      file_suffix='command_runner._clean_pod_spec.clean_spec'
    )
    clean_spec['containers'] = self._clean_containers_spec(
      clean_spec['containers']
    )
    return clean_spec

  def _image_name(self, tag_or_digest):
    ''' compose name for the docker image field in pod spec '''
    separator = '@' if ':' in tag_or_digest else ':'
    image_name = f'{self.ecr.project_repo}{separator}{tag_or_digest}'
    return image_name

  def _name(self):
    ''' fabricate a name for the pod and container '''
    name = '-'.join(
      filter(
        None,
        [
          f'{config.project_name}-hokusai-run',
          user(),
          k8s_uuid()
        ]
      )
    )
    return name

  def _overrides(self, cmd, constraint, env, tag_or_digest, pod_spec):
    ''' create overrides spec for kubectl '''
    overrides = { 'apiVersion': 'v1', 'spec': pod_spec}
    overrides['spec']['containers'][0]['args'] = cmd.split(' ')
    overrides['spec']['containers'][0]['name'] = self.container_name
    overrides['spec']['containers'][0]['image'] = self._image_name(
      tag_or_digest
    )
    constraint = constraint or config.run_constraints
    if constraint:
      overrides['spec']['nodeSelector'] = {}
      overrides['spec']['nodeSelector'].update(
        key_value_list_to_dict(constraint)
      )
    if env:
      overrides['spec']['containers'][0]['env'] = []
      for name, value in key_value_list_to_dict(env).items():
        overrides['spec']['containers'][0]['env'].append(
          {'name': name, 'value': value}
        )
    return overrides

  def _run_no_tty(self, cmd, image_name, overrides):
    ''' run command without tty '''
    args = ' '.join(
      [
        'run',
        self.pod_name,
        '--attach',
        f'--image={image_name}',
        f'--overrides={pipes.quote(json.dumps(overrides))}',
        '--restart=Never',
        '--rm'
      ]
    )
    file_debug(
      overrides,
      file_suffix='command_runner.run_no_tty.overrides'
    )
    return returncode(
      self.kctl.command(args)
    )

  def _run_tty(self, cmd, image_name, overrides):
    ''' run command with tty '''
    overrides['spec']['containers'][0].update({
      'stdin': True,
      'stdinOnce': True,
      'tty': True
    })
    args = ' '.join(
      [
        'run',
        self.pod_name,
        '-t',
        '-i',
        f'--image={image_name}',
        '--restart=Never',
        f'--overrides={pipes.quote(json.dumps(overrides))}',
        '--rm'
      ]
    )
    file_debug(
      overrides,
      file_suffix='command_runner.run_tty.overrides'
    )
    shout(
      self.kctl.command(args),
      print_output=True
    )

  def run(self, tag_or_digest, cmd, tty=None, env=(), constraint=()):
    '''
    spawn a pod to run the specified command,
    create an overrides spec to be given to kubectl run,
    base the override on:
    - pod spec of a model deployment spec in the appropriate hokusai yaml
    - the args to this function
    '''
    pod_spec = YamlSpec(
      self.yaml_template,
      render_template=True
    ).extract_pod_spec(self.model_deployment)
    file_debug(
      pod_spec,
      file_suffix='command_runner.run.pod_spec'
    )
    # clean specs so they are appropriate for kubectl run
    clean_pod_spec = self._clean_pod_spec(pod_spec)
    run_tty = tty if tty is not None else config.run_tty
    overrides = self._overrides(
      cmd,
      constraint,
      env,
      tag_or_digest,
      clean_pod_spec
    )
    image_name = self._image_name(tag_or_digest)
    if run_tty:
      self._run_tty(cmd, image_name, overrides)
    else:
      return self._run_no_tty(cmd, image_name, overrides)



import os
import sys

import json

from hokusai.lib.common import shout, write_temp_file
from hokusai.lib.config import config, HOKUSAI_TMP_DIR
from hokusai.lib.exceptions import HokusaiError
from hokusai.services.kubectl import Kubectl


class ConfigMap:
  ''' represent a Kubernetes ConfigMap '''
  def __init__(self, context, namespace='default', name=None):
    self.context = context
    self.kctl = Kubectl(context, namespace=namespace)
    self.name = name or f'{config.project_name}-environment'
    self.metadata = {
      'name': self.name,
      'namespace': namespace
    }
    if name is None:
      self.metadata['labels'] = {
        'app': config.project_name
      }
    self.struct = {
      'apiVersion': 'v1',
      'kind': 'ConfigMap',
      'metadata': self.metadata,
      'data': {}
    }

  def all(self):
    ''' return configmap data '''
    return self.struct['data']

  def create(self):
    ''' create configmap in Kubernetes '''
    path = write_temp_file(
      json.dumps(self.struct), HOKUSAI_TMP_DIR
    )
    self.kctl.create(path)

  def delete(self, key):
    ''' delete one key in configmap '''
    try:
      del self.struct['data'][key]
    except KeyError:
      raise HokusaiError(
        "Cannot unset '%s' as it does not exist" % key
      )

  def destroy(self):
    ''' delete configmap in Kubernetes '''
    shout(
      self.kctl.command("delete configmap %s" % self.name),
      print_output=True
    )

  def load(self):
    ''' read configmap data from Kubernetes '''
    payload = shout(
      self.kctl.command("get configmap %s -o json" % self.name)
    )
    struct = json.loads(payload)
    struct['data'] = struct.setdefault('data', {})
    self.struct['data'] = struct['data']

  def save(self):
    ''' save changes to Kubernetes '''
    path = write_temp_file(
      json.dumps(self.struct), HOKUSAI_TMP_DIR
    )
    self.kctl.apply(path)

  def update(self, key, value):
    ''' update value of a key in configmap '''
    self.struct['data'].update({key: value})



import os

import atexit
import yaml

from botocore.exceptions import NoCredentialsError

from hokusai.lib.common import (
  print_yellow,
  sorted_unique_list,
  unlink_file_if_not_debug,
  write_temp_file
)
from hokusai.lib.config import config, HOKUSAI_TMP_DIR
from hokusai.lib.config_loader import ConfigLoader
from hokusai.lib.exceptions import HokusaiError
from hokusai.lib.template_renderer import TemplateRenderer
from hokusai.services.ecr import ECR


class YamlSpec:
  ''' manage Hokusai yaml template '''
  def __init__(self, template_file, render_template=True):
    self.template_file = template_file
    self.ecr = ECR()
    self.tmp_filename = None
    self.render_template = render_template
    atexit.register(self.cleanup)

  def all_deployments_configmap_refs(self):
    ''' return list of unique configmaps referenced in deployments '''
    configmap_refs = []
    deployment_specs = self.get_resources_by_kind('Deployment')
    for spec in deployment_specs:
      configmap_refs += self.deployment_configmap_refs(spec)
    return sorted_unique_list(configmap_refs)

  def all_deployments_sa_refs(self):
    ''' return list of unique service accounts referenced in deployments '''
    sa_refs = []
    deployment_specs = self.get_resources_by_kind('Deployment')
    for spec in deployment_specs:
      sa_ref = self.deployment_sa_ref(spec)
      if sa_ref is not None:
        sa_refs += [sa_ref]
    return sorted_unique_list(sa_refs)

  def cleanup(self):
    unlink_file_if_not_debug(self.tmp_filename)

  def deployment_sa_ref(self, deployment_spec):
    ''' return any service account referenced in deployment spec '''
    pod_spec = deployment_spec['spec']['template']['spec']
    if 'serviceAccountName' in pod_spec:
      return pod_spec['serviceAccountName']
    else:
      return None

  def deployment_configmap_refs(self, deployment_spec):
    ''' return list of unique configmaps referenced in a deployment '''
    configmap_refs = []
    pod_spec = deployment_spec['spec']['template']['spec']
    if 'initContainers' in pod_spec:
      init_container_specs = pod_spec['initContainers']
      configmap_refs += self.containers_configmap_refs(init_container_specs)
    if 'containers' in pod_spec:
      container_specs = pod_spec['containers']
      configmap_refs += self.containers_configmap_refs(container_specs)
    return sorted_unique_list(configmap_refs)

  def containers_configmap_refs(self, container_specs):
    ''' return list of unique configmaps referenced in container specs '''
    configmap_refs = []
    for spec in container_specs:
      configmap_refs += self.container_configmap_refs(spec)
    return sorted_unique_list(configmap_refs)

  def container_configmap_refs(self, container_spec):
    ''' return list of configmaps referenced in a container spec '''
    configmap_refs = []
    if 'envFrom' in container_spec:
      for envfrom_spec in container_spec['envFrom']:
        if 'configMapRef' in envfrom_spec:
          configmap_refs += [
            envfrom_spec['configMapRef']['name']
          ]
    return sorted(configmap_refs)

  def extract_pod_spec(self, deployment_name):
    ''' extract pod spec from spec of specified deployment '''
    deployment_spec = self.get_resource_spec(
      'Deployment',
      deployment_name
    )
    if not deployment_spec['spec']['template']['spec']:
      raise HokusaiError(
        f'Pod spec in {deployment_name} deployment is empty.'
      )
    return deployment_spec['spec']['template']['spec']

  def get_resource_spec(self, kind, name=None):
    '''
    given 'kind' and 'metadata/name' of a Kubernetes resource,
    return its spec found in Hokusai yaml,
    if name is not specified,
    return the spec of the first resource matching kind
    '''
    right_kinds_spec = self.get_resources_by_kind(kind)
    for item in right_kinds_spec:
      if not name or item['metadata']['name'] == name:
        return item
    raise HokusaiError(
      f'Failed to find {name} {kind} resource in {self.template_file}'
    )

  def get_resources_by_kind(self, kind):
    ''' return specs of all resources found in Hokusai yaml matching kind '''
    spec = []
    yaml_spec = self.to_list()
    for item in yaml_spec:
      if item['kind'] == kind:
        spec += [item]
    return spec

  def to_file(self):
    ''' write rendered template to file '''
    path = write_temp_file(self.to_string(), HOKUSAI_TMP_DIR)
    self.tmp_filename = path
    return path

  def to_list(self):
    ''' convert rendered template to yaml list '''
    return list(yaml.safe_load_all(self.to_string()))

  def to_string(self):
    ''' render template file into string '''
    if self.render_template:
      template_config = {
        "project_name": config.project_name
      }

      try:
        template_config["project_repo"] = self.ecr.project_repo
      except NoCredentialsError:
        print_yellow(
          "WARNING: Could not get template variable project_repo"
        )

      if config.template_config_files:
        for template_config_file in config.template_config_files:
          try:
            config_loader = ConfigLoader(template_config_file)
            template_config.update(config_loader.load())
          except NoCredentialsError:
            print_yellow(
              f'WARNING: Could not get template config file {template_config_file}'
            )

      return TemplateRenderer(
        self.template_file, template_config
      ).render()
    else:
      with open(self.template_file, 'r') as f:
        content = f.read().strip()
      return content



from hokusai.services.command_runner import CommandRunner

def run(context, cmd, tty, tag, env, constraint, namespace=None):
  if tag is None:
    tag = context

  return CommandRunner(context, namespace=namespace).run(tag, cmd, tty=tty, env=env, constraint=constraint)



import yaml

from collections import OrderedDict

def represent_odict(dump, tag, mapping, flow_style=None):
    """Like BaseRepresenter.represent_mapping, but does not issue the sort().
    """
    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dump.alias_key is not None:
        dump.represented_objects[dump.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = list(mapping.items())
    for item_key, item_value in mapping:
        node_key = dump.represent_data(item_key)
        node_value = dump.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and not node_key.style):
            best_style = False
        if not (isinstance(node_value, yaml.ScalarNode) and not node_value.style):
            best_style = False
        value.append((node_key, node_value))
    if flow_style is None:
        if dump.default_flow_style is not None:
            node.flow_style = dump.default_flow_style
        else:
            node.flow_style = best_style
    return node

yaml.SafeDumper.add_representer(OrderedDict,
    lambda dumper, value: represent_odict(dumper, 'tag:yaml.org,2002:map', value))



import click

import hokusai

from hokusai.cli.base import base
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS
from hokusai.lib.config import config

@base.group()
def dev(context_settings=CONTEXT_SETTINGS):
  """Interact with docker-compose development environment
  defined by ./hokusai/development.yml"""
  pass


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.option('--build/--no-build', default=True, help='Force a build of the :latest image before starting (default: true)')
@click.option('-d', '--detach', type=click.BOOL, is_flag=True, help="Run containers in the background")
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def start(build, detach, filename, verbose):
  """Start the development environment defined in ./hokusai/development.yml"""
  set_verbosity(verbose)
  command(
    hokusai.dev_start,
    build,
    detach,
    filename
  )


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def stop(filename, verbose):
  """Stop the development environment defined in ./hokusai/development.yml"""
  set_verbosity(verbose)
  command(
    hokusai.dev_stop,
    filename
  )


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def status(filename, verbose):
  """Print the status of the development environment defined in ./hokusai/development.yml"""
  set_verbosity(verbose)
  command(
    hokusai.dev_status,
    filename
  )


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--follow', type=click.BOOL, is_flag=True, help="Follow output")
@click.option('-t', '--tail', type=click.INT, help="Number of lines of recent logs to display")
@click.option('-fi', '--filename', type=click.STRING, help='Use the docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def logs(follow, tail, filename, verbose):
  """Print logs from the development environment defined in ./hokusai/development.yml"""
  set_verbosity(verbose)
  command(
    hokusai.dev_logs,
    follow,
    tail,
    filename
  )


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.argument('container_command')
@click.option('--service-name', type=click.STRING, help="The service name to launch the container as (default: the name 'project-name' in `hokusai/config.yml`)")
@click.option('--stop', type=click.BOOL, is_flag=True, help='Stop all services after running the command')
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def run(container_command, service_name, stop, filename, verbose):
  """Run a command in a new container in the development environment defined in ./hokusai/development.yml"""
  set_verbosity(verbose)
  command(
    hokusai.dev_run,
    container_command,
    service_name,
    stop,
    filename
  )


@dev.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default ./hokusai/development.yml.j2)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def clean(filename, verbose):
  """Stop and remove all containers in the development environment defined in ./hokusai/development.yml.j2"""
  set_verbosity(verbose)
  command(
    hokusai.dev_clean,
    filename
  )



import os

from hokusai.lib.exceptions import HokusaiError

class TemplateSelector:
  def get(self, path):
    _path_root, _file_ext = os.path.splitext(path)

    if _file_ext:
      if not os.path.isfile(path):
        raise HokusaiError("File %s does not exist." % path)
      return path

    if os.path.isfile(path + '.yml.j2'):
      return path + '.yml.j2'
    if os.path.isfile(path + '.yaml.j2'):
      return path + '.yaml.j2'
    if os.path.isfile(path + '.yml'):
      return path + '.yml'
    if os.path.isfile(path + '.yaml'):
      return path + '.yaml'

    raise HokusaiError("No Yaml or Jinja templates found for %s" % os.path.basename(path))



# -*- coding: utf-8 -*-

import os

import click
import json
import platform
import random
import re
import shutil
import signal
import string
import tempfile

from botocore import session as botosession
from datetime import datetime
from pathlib import Path
from subprocess import (
  call, check_call, check_output, Popen, STDOUT
)
from tempfile import NamedTemporaryFile
from termcolor import cprint
from urllib.parse import urlparse

from hokusai.lib.constants import YAML_HEADER
from hokusai.lib.exceptions import (
  CalledProcessError, HokusaiError
)
from hokusai.services.s3 import S3Interface


CONTEXT_SETTINGS = {
  'terminal_width': 10000,
  'max_content_width': 10000,
  'help_option_names': ['-h', '--help']
}

EXIT_SIGNALS = [signal.SIGHUP, signal.SIGINT, signal.SIGQUIT, signal.SIGPIPE, signal.SIGTERM]

VERBOSE = False

AWS_DEFAULT_REGION = 'us-east-1'


def ansi_escape(raw_string):
  ''' rid string of ANSI esapes such as color '''
  escape_regex = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
  return escape_regex.sub('', raw_string)

def clean_dict(d1, fields_to_keep):
  ''' return dict ensuring that it contains only desired fields '''
  return {
    field: d1[field]
    for field in fields_to_keep
    if field in d1
  }

def clean_string(str):
  return str.lower().replace('_', '-')

def delete_keys(d1, keys):
  '''
  given dict 'd1',
  and dict 'keys' which specifies keys in d1 to delete,
  delete those keys from d1, if they exist.
  for example:

  # d1
  {
    'a': 1,
    'b': {
      'ba': 2,
      'bb': 3
    }
  }

  # keys
  {
    'a': {}, # empty dict because value doesn't matter
    'b': {
      'ba': {}
    },
    'c': {} # doesn't exist in d1
  }

  # d1 after key deletion
  {
    'b': {
      'bb': 3
    }
  }
  '''
  for key in keys:
    if key in d1:
      if keys[key] == {}:
        del d1[key]
      else:
        delete_keys(d1[key], keys[key])

def file_debug(dict1, file_suffix=None):
  ''' write dict to a file for debug '''
  # due to circular import
  from hokusai.lib.config import HOKUSAI_TMP_DIR
  if os.environ.get('DEBUG'):
    with NamedTemporaryFile(
      delete=False,
      dir=HOKUSAI_TMP_DIR,
      mode='w',
      suffix=file_suffix
    ) as temp_file:
      pretty_json = json.dumps(dict1, indent=2)
      temp_file.write(pretty_json)

def get_platform():
  ''' get the platform (e.g. darwin, linux) of the machine '''
  return platform.system().lower()

def get_region_name():
  # boto3 autodiscovery
  _region = botosession.get_session().get_config_variable('region')
  if _region:
    return _region
  # boto2 compatibility
  if os.environ.get('AWS_REGION'):
    return os.environ.get('AWS_REGION')
  return AWS_DEFAULT_REGION

def get_verbosity():
  global VERBOSE
  return VERBOSE

def k8s_uuid():
  uuid = []
  for i in range(0,5):
    uuid.append(random.choice(string.ascii_lowercase))
  return ''.join(uuid)

def key_value_list_to_dict(key_value_list):
  '''
  given:
  [
    'key1=value1',
    'key2=value2',
    ...
  ]

  return:
  {
    'key1': 'value1',
    'key2': 'value2',
    ...
  }
  '''
  splitted_key_value_list = (
    key_value.split('=', 1)
    for key_value in key_value_list
    if validate_key_value(key_value)
  )
  return {
    key: value
    for key, value in splitted_key_value_list
  }

def local_to_local(
  source_path,
  target_dir,
  target_file,
  create_target_dir=True,
  mode=0o660
):
  '''
  copy source file to target dir/file
  optionally create target dir, set target file mode as desired
  0o660 = rw-rw---- (read/write by file owner or group owner)
  '''
  # expand any ~
  expanded_target_dir = os.path.expanduser(target_dir)

  target_path = os.path.join(expanded_target_dir, target_file)
  target_path_obj = Path(target_path)

  # raise if target exists and is a dir
  if target_path_obj.is_dir():
    raise HokusaiError(f'Target {target_path} is an existing directory!')

  # if target exists and is a file, back it up
  if target_path_obj.is_file():
    backup_path = f'{target_path}.backup.{utc_yyyymmdd()}'
    shutil.copyfile(target_path, backup_path)

  # if target dir does not exist, create it if desired
  if create_target_dir:
    target_dir_obj = Path(expanded_target_dir)
    # noop if dir exists, error if what exists is a file
    target_dir_obj.mkdir(parents=True, exist_ok=True)

  shutil.copyfile(source_path, target_path)
  os.chmod(target_path, mode)

def pick_no():
  return random.choice([
    "Nope", "No", "нет", "Ne", "नहीं", "Daabi", "Nein", "Nay", "Nē", "ні", "خیر", "Nie", "Non", "ניט", "не", "아니", "いや", "没有", "Não"
  ])

def pick_yes():
  return random.choice([
    "Yep", "Si", "да", "Da", "Aane", "हाँ", "Ja", "はい", "Jā", "так", "بله", "Tak", "Wi", "Oui", "יאָ", "예", "是", "Sim"
  ])

def print_green(msg, newline_before=False, newline_after=False):
  cprint(smart_str(msg, newline_before, newline_after), 'green')

def print_red(msg, newline_before=False, newline_after=False):
  cprint(smart_str(msg, newline_before, newline_after), 'red')

def print_smart(msg, newline_before=False, newline_after=False):
  print(smart_str(msg, newline_before, newline_after))

def print_yellow(msg, newline_before=False, newline_after=False):
  cprint(smart_str(msg, newline_before, newline_after), 'yellow')

def returncode(command, mask=()):
  return call(verbose(command, mask=mask), stderr=STDOUT, shell=True)

def set_verbosity(v):
  # due to circular import
  from hokusai.lib.config import config
  global VERBOSE
  VERBOSE = v or config.always_verbose

def shout(command, print_output=False, mask=()):
  try:
    if print_output:
      return check_call(verbose(command, mask=mask), stderr=STDOUT, shell=True)
    else:
      retval = check_output(verbose(command, mask=mask), stderr=STDOUT, shell=True)
      if type(retval) == bytes:
        return retval.decode('utf-8')
      else:
        return retval

  except CalledProcessError as e:
    if mask:
      if hasattr(e, 'cmd') and e.cmd is not None:
        if type(e.cmd) == bytes:
          cmd = re.sub(mask[0], mask[1], e.cmd.decode('utf-8'))
        else:
          cmd = re.sub(mask[0], mask[1], e.cmd)
        e.cmd = cmd
      if hasattr(e, 'output') and e.output is not None:
        if type(e.output) == bytes:
          output = re.sub(mask[0], mask[1], e.output.decode('utf-8'))
        else:
          output = re.sub(mask[0], mask[1], e.output)
        e.output = output
    raise

def shout_concurrent(commands, print_output=False, mask=()):
  if print_output:
    processes = [Popen(verbose(command, mask=mask), shell=True) for command in commands]
  else:
    processes = [
      Popen(
        verbose(command, mask=mask),
        shell=True,
        stdout=open(os.devnull, 'w'),
        stderr=STDOUT
      ) for command in commands
    ]

  return_codes = []
  try:
    for p in processes:
      return_codes.append(p.wait())
  except KeyboardInterrupt:
    for p in processes:
      p.terminate()
    return_codes = [1 for p in processes]
  return return_codes

def smart_str(s, newline_before=False, newline_after=False):
  if newline_before:
    s = '\n' + s
  if newline_after:
    s = s + '\n'

  if isinstance(s, bytes):
    return s.decode('utf-8')
  if isinstance(s, int) or isinstance(s, float):
      return str(s)

  return s

def sorted_unique_list(list1):
  ''' return sorted list of unique elements in list1 '''
  return sorted(
    set(list1)
  )

def unlink_file_if_not_debug(path):
  ''' if DEBUG is off, unlink specified file '''
  if not os.environ.get('DEBUG'):
    try:
      os.unlink(path)
    except:
      pass

def uri_to_local(uri, target_dir, target_file):
  '''
  copy file from uri to target_dir/target_file
  support s3 scheme uri: s3://bucket/key
  support no scheme uri (i.e. local file): /path/to/file, ./path/to/file
  '''
  parsed_uri = urlparse(uri)
  tmpdir = tempfile.mkdtemp()
  try:
    if parsed_uri.scheme == 's3':
      s3_interface = S3Interface()
      filename = 'downloaded_file'
      file_path = os.path.join(tmpdir, filename)
      s3_interface.download(uri, file_path)
      local_to_local(file_path, target_dir, target_file)
    elif parsed_uri.scheme == '':
      local_to_local(parsed_uri.path, target_dir, target_file)
    else:
      raise HokusaiError(
        f'URI {uri} has unsupported scheme. Only "s3://" and file paths are supported.'
      )
  except:
    print_red(
      f'Error: failed to copy {uri} to {os.path.join(target_dir, target_file)}'
    )
    raise
  finally:
    shutil.rmtree(tmpdir)

def user():
  ''' obtain user name from environment '''
  user = None
  if os.environ.get('USER') is not None:
    # The regex used for the validation of name is
    # '[a-z0-9]([-a-z0-9]*[a-z0-9])?'
    user = re.sub(
      "[^0-9a-z]+", "-", os.environ.get('USER').lower()
    )
  return user

def utc_yyyymmdd():
  '''
  return yymmdd (utc), example:
  for January 15th, 1999, return 19990115
  '''
  return datetime.utcnow().strftime("%Y%m%d")

def validate_key_value(key_value):
  ''' return true if key_value is in the form KEY=VALUE, else raise '''
  if '=' not in key_value:
    raise HokusaiError(
      "Error: key/value pair must be of the form 'KEY=VALUE'"
    )
  return True

def verbose(msg, mask=()):
  if VERBOSE:
    if mask:
      print_yellow(
        "==> hokusai exec `%s`" % re.sub(mask[0], mask[1], msg),
        newline_after=True
      )
    else:
      print_yellow("==> hokusai exec `%s`" % msg, newline_after=True)
  return msg

def write_temp_file(data, dir):
  ''' write data to temp file, return file path '''
  obj = NamedTemporaryFile(delete=False, dir=dir, mode='w')
  obj.write(data)
  obj.close()
  return obj.name

def yaml_content_with_header(content_string):
  ''' return content with yaml header prepended '''
  return YAML_HEADER + content_string

def confirm_delete_resources(project_name, environment, yaml_file):
  ''' Display confirmation prompt for destructive delete operations '''
  print_smart("\n" + "="*60)
  print_red("⚠️  DESTRUCTIVE ACTION WARNING", newline_before=False, newline_after=False)
  print_smart("="*60)
  
  print_smart(f"Project:     {project_name}")
  print_smart(f"Environment: {environment}")
  if yaml_file is not None:
    print_smart(f"Config file: {yaml_file}")
  print_smart("")
  print_yellow("This will permanently delete Kubernetes resources", newline_before=False, newline_after=False)
  print_yellow("for this project in the specified environment.", newline_before=False, newline_after=False)
  print_smart("")
  print_smart("This action CANNOT be undone!")
  print_smart("="*60)
  
  if not click.confirm("Do you want to continue with the deletion?"):
    print_smart("Deletion cancelled.")
    return False
    
  return True



import os
import signal

from hokusai.lib.common import EXIT_SIGNALS, print_green, print_red, shout
from hokusai.lib.config import TEST_YML_FILE, config
from hokusai.lib.docker_compose_helpers import generate_compose_command, get_yaml_template
from hokusai.lib.exceptions import CalledProcessError, HokusaiError
from hokusai.services.docker import Docker


def test(build, cleanup, filename, service_name):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=TEST_YML_FILE
  )

  def on_cleanup(*args):
    shout(
      f'{compose_command} -p hokusai stop',
    )
    shout(
      f'{compose_command} -p hokusai rm --force',
    )
  if cleanup:
    for sig in EXIT_SIGNALS:
      signal.signal(sig, on_cleanup)

  if build:
    yaml_template = get_yaml_template(
      filename,
      default_yaml_file=TEST_YML_FILE
    )
    Docker().build(filename=yaml_template)

  if service_name is None:
    service_name = config.project_name

  opts = " --abort-on-container-exit --exit-code-from %s" % service_name
  print_green(
    "Starting test environment... Press Ctrl+C to stop.",
    newline_after=True
  )
  try:
    return_code = int(
      shout(
        f'{compose_command} -p hokusai up{opts}',
        print_output=True
      )
    )
  except CalledProcessError as e:
    print_red("CalledProcessError return code: %s" % e.returncode)
    if cleanup: on_cleanup()
    raise HokusaiError('Tests Failed')

  if return_code:
    raise HokusaiError(
      'Tests Failed - Exit Code: %s\n' % return_code,
      return_code=return_code
    )
  else:
    print_green("Tests Passed")

  if cleanup: on_cleanup()

  return return_code



import os
import click

import hokusai

from hokusai import CWD
from hokusai.cli.base import base
from hokusai.cli.staging import KUBE_CONTEXT
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS, clean_string, print_yellow
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, config


@base.group('review_app')
def review_app(context_settings=CONTEXT_SETTINGS):
  """Create/Manage review apps"""
  pass

@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
@click.option(
  '-sf',
  '--source-file',
  type=click.STRING,
  default='hokusai/staging.yml',
  help="The source yaml file from which to create the new resource file (default: hokusai/staging.yml)"
)
def create_yaml(app_name, verbose, source_file):
  """Create Review App Yaml"""
  set_verbosity(verbose)
  command(
    hokusai.create_yaml_for_cli,
    source_file,
    app_name
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
@click.option(
  '-sf',
  '--source-file',
  type=click.STRING,
  default='hokusai/staging.yml',
  help="The source yaml file from which to create the new resource file (default: hokusai/staging.yml)"
)
def setup(app_name, verbose, source_file):
  """Setup a new Review App"""
  set_verbosity(verbose)
  command(
    hokusai.setup_review_app,
    source_file,
    app_name
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def create(app_name, verbose):
  """Creates the Kubernetes based resources defined in ./hokusai/{APP_NAME}.yml"""
  command(
    hokusai.k8s_create,
    KUBE_CONTEXT,
    tag=app_name,
    namespace=clean_string(app_name),
    filename=os.path.join(CWD, HOKUSAI_CONFIG_DIR, "%s.yml" % app_name),
    render_template=False
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def delete(app_name, verbose):
  """Deletes the Kubernetes based resources defined in ./hokusai/{APP_NAME}.yml"""
  set_verbosity(verbose)
  command(
    hokusai.delete_review_app,
    KUBE_CONTEXT,
    app_name,
    filename=os.path.join(CWD, HOKUSAI_CONFIG_DIR, "%s.yml" % app_name),
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def update(app_name, verbose):
  """Updates the Kubernetes based resources defined in ./hokusai/{APP_NAME}.yml"""
  set_verbosity(verbose)
  command(
    hokusai.k8s_update,
    KUBE_CONTEXT,
    namespace=clean_string(app_name),
    filename=os.path.join(CWD, HOKUSAI_CONFIG_DIR, "%s.yml" % app_name),
    skip_checks=True,
    render_template=False
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option(
  '--resources/--no-resources',
  default=True,
  help='Print Kubernetes API objects defined in ./hokusai/APP_NAME.yml (default: true)'
)
@click.option('--pods/--no-pods', default=True, help='Print pods (default: true)')
@click.option('--describe', type=click.BOOL, is_flag=True, help="Print 'kubectl describe' output")
@click.option('--top', type=click.BOOL, is_flag=True, help='Print top pods')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def status(app_name, resources, pods, describe, top, verbose):
  """Print the Kubernetes resources status defined in the staging context / {APP_NAME} namespace"""
  set_verbosity(verbose)
  command(
    hokusai.k8s_status,
    KUBE_CONTEXT,
    resources,
    pods,
    describe,
    top,
    namespace=clean_string(app_name),
    filename=os.path.join(CWD, HOKUSAI_CONFIG_DIR, "%s.yml" % app_name),
    render_template=False
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.argument('container_command', type=click.STRING)
@click.option('--tty', type=click.BOOL, is_flag=True, help='Attach the terminal')
@click.option('--tag', type=click.STRING, help='The image tag to run (defaults to APP_NAME)')
@click.option('--env', type=click.STRING, multiple=True, help='Environment variables in the form of "KEY=VALUE"')
@click.option(
  '--constraint',
  type=click.STRING,
  multiple=True,
  help='Constrain command to run on nodes matching labels in the form of "key=value"'
)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def run(app_name, container_command, tty, tag, env, constraint, verbose):
  """Launch a new container and run a command"""
  set_verbosity(verbose)
  if tag is None:
    tag = clean_string(app_name)
  command(
    hokusai.run,
    KUBE_CONTEXT,
    container_command,
    tty,
    tag,
    env,
    constraint,
    namespace=clean_string(app_name)
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-s', '--timestamps', type=click.BOOL, is_flag=True, help='Include timestamps')
@click.option('-f', '--follow', type=click.BOOL, is_flag=True, help='Follow logs')
@click.option('-t', '--tail', type=click.INT, help="Number of lines of recent logs to display")
@click.option('-p', '--previous', type=click.BOOL, is_flag=True, help='Get from the previous run of the Pod container(s)')
@click.option(
  '-l',
  '--label',
  type=click.STRING,
  multiple=True,
  help='Filter pods by additional label selectors in the form of "key=value"'
)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def logs(app_name, timestamps, follow, tail, previous, label, verbose):
  """Get container logs"""
  set_verbosity(verbose)
  command(
    hokusai.logs,
    KUBE_CONTEXT,
    timestamps,
    follow,
    tail,
    previous,
    label,
    namespace=clean_string(app_name)
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.argument('tag', type=click.STRING)
@click.option('--migration', type=click.STRING, help='Run a migration before deploying')
@click.option(
  '--constraint',
  type=click.STRING,
  multiple=True,
  help='Constrain migration and deploy hooks to run on nodes matching labels in the form of "key=value"'
)
@click.option('--git-remote', type=click.STRING, help='Push deployment tags to git remote')
@click.option(
  '-t',
  '--timeout',
  type=click.INT,
  default=600,
  help="Timeout deployment rollout after N seconds (default 600)"
)
@click.option('-u', '--update-config', type=click.BOOL, is_flag=True, help='Also update Kubernetes config')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def deploy(app_name, tag, migration, constraint, git_remote, timeout, update_config, verbose):
  """Update the project's deployment(s) to reference the given image tag"""
  set_verbosity(verbose)
  command(
    hokusai.update,
    KUBE_CONTEXT,
    tag,
    migration,
    constraint,
    git_remote,
    timeout,
    namespace=clean_string(app_name),
    update_config=update_config,
    filename=os.path.join(CWD, HOKUSAI_CONFIG_DIR, "%s.yml" % app_name),
    render_template=False
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def list(verbose):
  """List existing review apps of the project"""
  set_verbosity(verbose)
  hokusai.list_review_apps(KUBE_CONTEXT, labels=f'app-name={config.project_name},app-phase=review')


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-d', '--deployment', type=click.STRING, help='Only refresh the given deployment')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def refresh(app_name, deployment, verbose):
  """Refresh the project's deployment(s) by recreating the currently running containers"""
  set_verbosity(verbose)
  command(
    hokusai.refresh,
    KUBE_CONTEXT,
    deployment,
    namespace=clean_string(app_name)
  )


@review_app.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option('-d', '--deployment', type=click.STRING, help='Only refresh the given deployment')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def restart(app_name, deployment, verbose):
  """Alias for 'refresh'"""
  set_verbosity(verbose)
  command(
    hokusai.refresh,
    KUBE_CONTEXT,
    deployment,
    namespace=clean_string(app_name)
  )


@review_app.group()
def env(context_settings=CONTEXT_SETTINGS):
  """Interact with the runtime environment for the review app"""
  pass


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.option(
  '--configmap',
  type=click.STRING,
  help="Copy the given ConfigMap (default: the project's environment ConfigMap)"
)
@click.option(
  '-v',
  '--verbose',
  type=click.BOOL,
  is_flag=True,
  help='Verbose output'
)
def copy(app_name, configmap, verbose):
  """Copies the app's environment config map to the namespace {APP_NAME}"""
  set_verbosity(verbose)
  print_yellow(
    'This command is no longer necessary. It is deprecated and will be removed soon. Please stop using it.'
  )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def get(app_name, env_vars, verbose):
  """Print environment variables stored on the Kubernetes server"""
  set_verbosity(verbose)
  command(
    hokusai.get_env,
    KUBE_CONTEXT,
    env_vars,
    namespace=clean_string(app_name)
  )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def set(app_name, env_vars, verbose):
  """Set environment variables - each of {ENV_VARS} must be in of form 'KEY=VALUE'"""
  set_verbosity(verbose)
  command(
    hokusai.set_env,
    KUBE_CONTEXT,
    env_vars,
    namespace=clean_string(app_name)
  )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_name', type=click.STRING)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def unset(app_name, env_vars, verbose):
  """Unset environment variables - each of {ENV_VARS} must be of the form 'KEY'"""
  set_verbosity(verbose)
  command(
    hokusai.unset_env,
    KUBE_CONTEXT,
    env_vars,
    namespace=clean_string(app_name)
  )



import os
import tempfile
import yaml

from shutil import rmtree

from hokusai.lib.common import print_red, uri_to_local
from hokusai.lib.exceptions import HokusaiError


class ConfigLoader:
  ''' load Hokusai config from Yaml files '''
  def __init__(self, uri):
    self.uri = uri

  def _load_from_file(self, file_path):
    ''' load config from file '''
    with open(file_path, 'r') as f:
      struct = yaml.safe_load(f.read())
      return struct

  def load(self):
    ''' load config from uri '''
    if not self.uri.endswith('yaml') and not self.uri.endswith('yml'):
      raise HokusaiError('Config file must be of Yaml file type')

    tmpdir = tempfile.mkdtemp()
    filename = 'hokusai.yml'
    tmp_configfile = os.path.join(tmpdir, filename)

    try:
      uri_to_local(self.uri, tmpdir, filename)
      config = self._load_from_file(tmp_configfile)
      return config
    except:
      print_red(f'Error: Not able to load config from {self.uri}')
      raise
    finally:
      rmtree(tmpdir)



from hokusai.cli.base import base
from hokusai.cli.dev import dev
from hokusai.cli.registry import registry
from hokusai.cli.staging import staging
from hokusai.cli.production import production
from hokusai.cli.pipeline import pipeline
from hokusai.cli.review_app import review_app



import os
import shutil
import tempfile

from urllib.request import urlretrieve

from hokusai.lib.common import (
  get_platform,
  local_to_local,
  print_green,
  print_red,
  uri_to_local
)
from hokusai.lib.global_config import HokusaiGlobalConfig


def install_kubectl(kubectl_version, kubectl_dir):
  ''' download and install kubectl '''
  tmpdir = tempfile.mkdtemp()
  try:
    url = (
      f'https://storage.googleapis.com/kubernetes-release/release/v' +
      f'{kubectl_version}' +
      f'/bin/{get_platform()}/amd64/kubectl'
    )
    download_to = os.path.join(tmpdir, 'kubectl')
    print_green(
      f'Downloading kubectl from {url} ...', newline_after=True
    )
    urlretrieve(url, download_to)
    print_green(
      f'Installing kubectl into {kubectl_dir} ...', newline_after=True
    )
    local_to_local(download_to, kubectl_dir, 'kubectl', mode=0o770)
  except:
    print_red(f'Error: Failed to install kubectl')
    raise
  finally:
    shutil.rmtree(tmpdir)

def install(global_config, skip_kubeconfig, skip_kubectl):
  ''' install kubeconfig and kubectl '''
  if not skip_kubeconfig:
    print_green(
      f'Downloading kubeconfig from ' +
      f'{global_config.kubeconfig_source_uri} ' +
      f'to {global_config.kubeconfig_dir} ...',
      newline_after=True
    )
    uri_to_local(
      global_config.kubeconfig_source_uri,
      global_config.kubeconfig_dir,
      'config'
    )
  if not skip_kubectl:
    install_kubectl(
      global_config.kubectl_version,
      global_config.kubectl_dir
    )

def hokusai_configure(
  kubeconfig_dir,
  kubectl_dir,
  skip_kubeconfig,
  skip_kubectl
):
  '''
  read new global config,
  save global config,
  install kubeconfig and kubectl
  '''
  global_config = HokusaiGlobalConfig()
  # override global config with cmdline options
  global_config.merge(
    kubectl_dir=kubectl_dir,
    kubeconfig_dir=kubeconfig_dir,
  )
  global_config.save()
  install(global_config, skip_kubeconfig, skip_kubectl)



import base64
import boto3
import re

from botocore.exceptions import BotoCoreError, ClientError

from hokusai.lib.common import get_region_name
from hokusai.lib.config import config
from hokusai.lib.exceptions import HokusaiError

SHA1_REGEX = re.compile(r"\b[0-9a-f]{40}\b")

class ECR:
  def __init__(self):
    self.client = boto3.client('ecr', region_name=get_region_name())
    self.__aws_account_id = None
    self.__registry = None
    self.__project_repo = None
    self.__images = None

  @property
  def aws_account_id(self):
    if self.__aws_account_id is None:
      self.__aws_account_id = boto3.client('sts', region_name=get_region_name()).get_caller_identity().get('Account')
    return self.__aws_account_id

  @property
  def registry(self):
    if self.__registry is None:
      repos = []
      res = self.client.describe_repositories(registryId=self.aws_account_id)
      try:
        repos += res['repositories']
      except KeyError as err:
        raise HokusaiError("Fetching ECR registry failed with error %s" % str(err))
      while 'nextToken' in res:
        res = self.client.describe_repositories(registryId=self.aws_account_id,
                                                  nextToken=res['nextToken'])
        repos += res['repositories']
      self.__registry = repos
    return self.__registry

  @property
  def images(self):
    if self.__images is None:
      images = []
      res = self.client.describe_images(registryId=self.aws_account_id,
                                    repositoryName=config.project_name)
      images += res['imageDetails']
      while 'nextToken' in res:
        res = self.client.describe_images(registryId=self.aws_account_id,
                                      repositoryName=config.project_name,
                                      nextToken=res['nextToken'])
        images += res['imageDetails']
      self.__images = images
    return self.__images

  @property
  def project_repo(self):
    if self.__project_repo is None:
      for repo in self.registry:
        if repo['repositoryName'] == config.project_name:
          self.__project_repo = repo['repositoryUri']
    return self.__project_repo

  def project_repo_exists(self):
    return self.project_repo is not None

  def create_project_repo(self):
    self.client.create_repository(repositoryName=config.project_name)
    self.__registry = None
    return True

  def get_login(self):
    res = self.client.get_authorization_token(registryIds=[str(self.aws_account_id)])['authorizationData'][0]
    token = base64.b64decode(res['authorizationToken']).decode('utf8')
    username = token.split(':')[0]
    password = token.split(':')[1]
    return "docker login -u %s -p %s %s" % (username, password, res['proxyEndpoint'])

  def get_image_by_tag(self, tag):
    try:
      res = self.client.describe_images(registryId=self.aws_account_id,
                                        repositoryName=config.project_name,
                                        imageIds=[{'imageTag': tag}])
    except self.client.exceptions.ImageNotFoundException:
      return
    if 'imageDetails' in res:
      return res['imageDetails'][0]

  def tags(self):
    tgs = []
    for image in self.images:
      if 'imageTags' not in list(image.keys()):
        continue
      for tag in image['imageTags']:
        tgs.append(tag)
    return tgs

  def deployment_tags(self, context):
    context_re = re.compile(r"%s--\d\d\d\d-\d\d-\d\d--\d\d\-\d\d-\d\d" % context)
    return sorted([x for x in self.tags() if context_re.match(x)])

  def current_deployment_tag(self, context):
    context_re = re.compile(r"%s--\d\d\d\d-\d\d-\d\d--\d\d\-\d\d-\d\d" % context)
    image = self.get_image_by_tag(context)
    if image is None:
      return
    for tag in image['imageTags']:
      if context_re.match(tag):
        return tag

  def tag_exists(self, tag):
    for image in self.images:
      if 'imageTags' not in list(image.keys()):
        continue
      if tag in image['imageTags']:
        return True
    return False

  def find_git_sha1_image_tag(self, tag):
    image = self.get_image_by_tag(tag)
    if image is None:
      return
    for t in image['imageTags']:
      if SHA1_REGEX.match(t) is not None:
        return t

  def image_digest_for_tag(self, tag):
    image = self.get_image_by_tag(tag)
    if image is None:
      return
    return image['imageDigest']

  def retag(self, tag, new_tag):
    res = self.client.batch_get_image(
      registryId=self.aws_account_id,
      repositoryName=config.project_name,
      imageIds=[{'imageTag': tag}]
    )

    if res['failures'] and not res['images']:
      raise ValueError("Failed to retrieve image manifest for tag %s" % tag)

    image = res['images'][0]

    try:
      self.client.put_image(
          registryId=self.aws_account_id,
          repositoryName=config.project_name,
          imageManifest=image['imageManifest'],
          imageTag=new_tag
      )
    except ClientError as e:
      if e.response['Error']['Code'] != 'ImageAlreadyExistsException':
        raise



import os
import sys
import shutil
import atexit

from collections import OrderedDict

import yaml

from packaging.specifiers import SpecifierSet, InvalidSpecifier
from packaging.version import Version, InvalidVersion

from hokusai import CWD
from hokusai.lib.common import yaml_content_with_header
from hokusai.lib.exceptions import HokusaiError
from hokusai.version import VERSION


HOKUSAI_ENV_VAR_PREFIX = 'HOKUSAI_'
HOKUSAI_CONFIG_DIR = 'hokusai'
HOKUSAI_TMP_DIR = os.path.join(CWD, '.hokusai-tmp')
HOKUSAI_CONFIG_FILE = os.path.join(CWD, HOKUSAI_CONFIG_DIR, 'config.yml')
BUILD_YAML_FILE = 'build'
TEST_YML_FILE = 'test'
DEVELOPMENT_YML_FILE = 'development'

class HokusaiConfig:
  def __init__(self):
    if not os.path.isdir(HOKUSAI_TMP_DIR):
      os.mkdir(HOKUSAI_TMP_DIR)

  def create(self, project_name):
    config = OrderedDict([
      ('project-name', project_name)
    ])

    with open(HOKUSAI_CONFIG_FILE, 'w') as f:
      payload = yaml_content_with_header(
        yaml.safe_dump(config, default_flow_style=False)
      )
      f.write(payload)

  def check(self):
    if not self._check_config_present(HOKUSAI_CONFIG_FILE):
      raise HokusaiError("Hokusai is not set up for this project - run 'hokusai setup'")
    if not self._check_required_version(self.hokusai_required_version, VERSION):
      raise HokusaiError("Hokusai's current version %s does not satisfy this project's version requirements '%s'.  Aborting."
                           % (VERSION, self.hokusai_required_version))

  def _check_config_present(self, config_file):
    return os.path.isfile(config_file)

  def _check_required_version(self, required_version, target_version):
    if required_version is None:
      return True
    try:
      match_versions = SpecifierSet(required_version, prereleases=True)
    except InvalidSpecifier:
      raise HokusaiError("Could not parse '%s' as a valid version specifier. See https://www.python.org/dev/peps/pep-0440/#version-specifiers" % required_version)
    try:
      compare_version = Version(target_version)
    except InvalidVersion:
      raise HokusaiError("Could not parse '%s' as a valid version identifier. See https://www.python.org/dev/peps/pep-0440/#version-scheme" % target_version)
    return compare_version in match_versions

  def get(self, key, default=None, use_env=False, _type=str):
    value = self._config_value_for(key, _type)
    if value is not None:
      return value

    if use_env:
      value = self._env_value_for(key, _type)
      if value is not None:
        return value

    return default

  def _env_value_for(self, key, _type):
    env_var = HOKUSAI_ENV_VAR_PREFIX + key.upper().replace('-', '_')
    val = os.environ.get(env_var)
    if val is None:
      return val
    if _type == list:
      try:
        return _type(val.split(','))
      except ValueError:
        raise HokusaiError("Environment variable %s could not be split to %s" % (env_var, _type))
    try:
      return _type(val)
    except ValueError:
      raise HokusaiError("Environment variable %s could not be cast to %s" % (env_var, _type))

  def _config_value_for(self, key, _type):
    try:
      with open(HOKUSAI_CONFIG_FILE, 'r') as config_file:
        config_struct = yaml.safe_load(config_file.read())
        try:
          val = config_struct[key]
        except KeyError:
          return None
        if not isinstance(val, _type):
          raise HokusaiError("Config key %s is not of %s" % (key, _type))
        return val
    except IOError:
      return None


  @property
  def always_verbose(self):
    return self.get('always-verbose', default=False, use_env=True, _type=bool)

  @property
  def follow_logs(self):
    return self.get('follow-logs', default=False, use_env=True, _type=bool)

  @property
  def git_remote(self):
    return self.get('git-remote')

  @property
  def hokusai_required_version(self):
    return self.get('hokusai-required-version')

  @property
  def post_build(self):
    return self.get('post-build')

  @property
  def post_deploy(self):
    return self.get('post-deploy')

  @property
  def pre_build(self):
    return self.get('pre-build')

  @property
  def pre_deploy(self):
    return self.get('pre-deploy')

  @property
  def project_name(self):
    project = self.get('project-name')
    if project is None:
      raise HokusaiError("Unconfigured 'project-name'! Plz check ./hokusai/config.yml")
    return project

  @property
  def run_constraints(self):
    return self.get('run-constraints', default=[], use_env=True, _type=list)

  @property
  def run_template(self):
    return self.get('run-template')

  @property
  def run_tty(self):
    return self.get('run-tty', default=False, use_env=True, _type=bool)

  @property
  def tail_logs(self):
    return self.get('tail-logs', use_env=True, _type=int)

  @property
  def template_config_files(self):
    return self.get('template-config-files', _type=list)

config = HokusaiConfig()



import os
import signal

from hokusai.lib.common import EXIT_SIGNALS, print_green, shout
from hokusai.lib.config import DEVELOPMENT_YML_FILE, config
from hokusai.lib.docker_compose_helpers import generate_compose_command, get_yaml_template
from hokusai.services.docker import Docker


def dev_start(build, detach, filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  def cleanup(*args):
    shout(
      f'{compose_command} -p hokusai stop',
      print_output=True
    )
  for sig in EXIT_SIGNALS:
    signal.signal(sig, cleanup)
  opts = ''
  if build:
    yaml_template = get_yaml_template(
      filename,
      default_yaml_file=DEVELOPMENT_YML_FILE
    )
    Docker().build(filename=yaml_template)
  if detach:
    opts += ' -d'
  if not detach:
    print_green(
      "Starting development environment... Press Ctrl+C to stop."
    )
  shout(
    f'{compose_command} -p hokusai up{opts}',
    print_output=True
  )
  if detach:
    print_green(
      "Run `hokousai dev stop` to shut down, or `hokusai dev logs --follow` to tail output."
    )

def dev_stop(filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  shout(
    f'{compose_command} -p hokusai stop',
    print_output=True
  )

def dev_status(filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  shout(
    f'{compose_command} -p hokusai ps',
    print_output=True
  )

def dev_logs(follow, tail, filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  opts = ''
  if follow:
    opts += ' --follow'
  if tail:
    opts += " --tail=%i" % tail
  shout(
    f'{compose_command} -p hokusai logs{opts}',
    print_output=True
  )

def dev_run(container_command, service_name, stop, filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  if service_name is None:
    service_name = config.project_name
  shout(
    f'{compose_command} -p hokusai run {service_name} {container_command}',
    print_output=True
  )
  if stop:
    shout(
      f'{compose_command} -p hokusai stop',
      print_output=True
    )

def dev_clean(filename):
  compose_command = generate_compose_command(
    filename,
    default_yaml_file=DEVELOPMENT_YML_FILE
  )
  shout(
    f'{compose_command} -p hokusai stop',
    print_output=True
  )
  shout(
    f'{compose_command} -p hokusai rm --force',
    print_output=True
  )



from hokusai.lib.common import print_red, print_green, shout
from hokusai.lib.config import config
from hokusai.services.deployment import Deployment
from hokusai.services.ecr import ECR
from hokusai.lib.exceptions import HokusaiError

def gitcompare(org_name, git_compare_link):
  ecr = ECR()

  staging_tag = ecr.find_git_sha1_image_tag('staging')
  if staging_tag is None:
    raise HokusaiError("Could not find a tag for staging.  Aborting.")

  production_tag = ecr.find_git_sha1_image_tag('production')
  if production_tag is None:
    raise HokusaiError("Could not find a git SHA1 tag for production.  Aborting.")

  print_green(git_compare_link % (org_name, config.project_name, production_tag, staging_tag))



from operator import itemgetter

from hokusai.services.ecr import ECR
from hokusai.lib.config import config
from hokusai.lib.common import print_green, print_yellow, print_red, print_smart, shout, pick_yes, pick_no

def images(tag_exists, reverse_sort, limit, filter_tags, digests):
  ecr = ECR()
  if tag_exists:
    if ecr.get_image_by_tag(tag_exists):
      print_green(pick_yes())
      return
    else:
      print_red(pick_no())
      return 1

  images = ecr.images
  sorted_images = sorted(images, key=itemgetter('imagePushedAt'), reverse=not reverse_sort)
  filtered_images = [image for image in sorted_images if 'imageTags' in list(image.keys())]
  if filter_tags:
    filtered_images = [image for image in filtered_images if filter_tags in ', '.join(image['imageTags'])]

  if digests:
    print_green('Image Pushed At           | Image Digest                                                            | Image Tags', newline_before=True)
    print_green('--------------------------------------------------------------------------------------------------------------------------------------')
  else:
    print_green('Image Pushed At           | Image Tags', newline_before=True)
    print_green('----------------------------------------------------------')

  for image in filtered_images[:limit]:
    image_tags = ', '.join(image['imageTags'])
    if digests:
      line = "%s | %s | %s" % (image['imagePushedAt'], image['imageDigest'], image_tags)
    else:
      line = "%s | %s" % (image['imagePushedAt'], image_tags)
    if 'production' in image['imageTags']:
      print_green(line)
    elif 'staging' in image['imageTags']:
      print_yellow(line)
    else:
      print_smart(line)

  print_yellow("%d more images available" % (len(filtered_images) - len(filtered_images[:limit])),
                newline_before=True, newline_after=True)




from hokusai.lib.config import config
from hokusai.services.ecr import ECR
from hokusai.lib.common import print_green, print_yellow, shout
from hokusai.lib.exceptions import HokusaiError
from hokusai.services.docker import Docker

def build_and_push(remote_tag, local_tag, filename, ecr, skip_latest=False):
  """Build Docker image and push it to ECR"""
  Docker().build(filename)
  push_only(remote_tag, local_tag, ecr, skip_latest)

def ecr_check():
  """Check that ECR has a repo for the project"""
  ecr = ECR()
  if not ecr.project_repo_exists():
    raise HokusaiError("ECR repo %s does not exist. Has 'hokusai setup' been run for this project?" % config.project_name)
  return ecr

def git_status_check():
  """Check working directory's Git status"""
  git_status_files = shout('git status --porcelain --ignored')
  if git_status_files:
    print_yellow("The following files/directories are either changed, untracked, or Git ignored.")
    print_yellow("If they are not excluded by .dockerignore, they may get copied into the resulting Docker image.")
    print_yellow(git_status_files)
    raise HokusaiError("Working directory unclean. Aborting. Use '--force' to force.")

def push_image(remote_tag, local_tag, build, filename, force, overwrite, skip_latest=False):
  """Push Docker image to ECR"""
  force or git_status_check()
  ecr = ecr_check()
  remote_tag = remote_tag_check(remote_tag, overwrite, ecr)
  if build:
    build_and_push(remote_tag, local_tag, filename, ecr, skip_latest)
  else:
    push_only(remote_tag, local_tag, ecr, skip_latest)

def push_only(remote_tag, local_tag, ecr, skip_latest=False):
  """Push Docker Compose built image to ECR"""
  shout(ecr.get_login(), mask=(r'^(docker login -u) .+ (-p) .+ (.+)$', r'\1 ****** \2 ***** \3'))
  docker_compose_repo = "hokusai_" + config.project_name
  tag_and_push(docker_compose_repo, local_tag, ecr.project_repo, remote_tag)
  if not skip_latest:
    tag_and_push(docker_compose_repo, local_tag, ecr.project_repo, 'latest')

def remote_tag_check(remote_tag, overwrite, ecr):
  """Set the tag to be used in ECR and see whether it already exists there"""
  if remote_tag is None:
    remote_tag = shout('git rev-parse HEAD').strip()
  if not overwrite and ecr.tag_exists(remote_tag):
    raise HokusaiError("Tag %s already exists in registry. Aborting. Use '--overwrite' to overwrite registry's tag." % remote_tag)
  return remote_tag

def tag_and_push(local_repo, local_tag, remote_repo, remote_tag):
  """Push local Docker repo:tag to remote repo:tag"""
  shout("docker tag %s:%s %s:%s" % (local_repo, local_tag, remote_repo, remote_tag))
  shout("docker push %s:%s" % (remote_repo, remote_tag), print_output=True)
  print_green("Pushed %s:%s to %s:%s" % (local_repo, local_tag, remote_repo, remote_tag), newline_after=True)



import os
import sys
import time
import tempfile

from shutil import rmtree, copyfile
from distutils.dir_util import mkpath

import yaml

from jinja2 import Environment, PackageLoader, FileSystemLoader, StrictUndefined

from hokusai import CWD
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, config
from hokusai.services.ecr import ECR
from hokusai.lib.common import print_green, clean_string, shout
from hokusai.lib.environment import templates_dir_path
from hokusai.lib.exceptions import HokusaiError

def setup(project_name, template_remote, template_dir, template_vars, allow_missing_vars):
  mkpath(os.path.join(CWD, HOKUSAI_CONFIG_DIR))
  config.create(clean_string(project_name))

  ecr = ECR()
  if ecr.project_repo_exists():
    print_green("Project repo %s already exists.  Skipping create." % ecr.project_repo)
  else:
    ecr.create_project_repo()
    print_green("Created project repo %s" % ecr.project_repo)

  scratch_dir = None
  if template_remote:
    scratch_dir = tempfile.mkdtemp()
    git_repo_and_branch = template_remote.split('#', 1)
    git_repo = git_repo_and_branch[0]
    if len(git_repo_and_branch) == 2:
      git_branch = git_repo_and_branch[1]
    else:
      git_branch = "main"
    shout("git clone -b %s --single-branch %s %s" % (git_branch, git_repo, scratch_dir))

  custom_template_dir = None
  if allow_missing_vars:
    environment_kwargs = {}
  else:
    environment_kwargs = { "undefined": StrictUndefined }

  if scratch_dir and template_dir:
    custom_template_dir = os.path.join(scratch_dir, os.path.basename(template_dir))
    env = Environment(loader=FileSystemLoader(os.path.join(scratch_dir, os.path.basename(template_dir))), **environment_kwargs)
  elif scratch_dir:
    custom_template_dir = scratch_dir
    env = Environment(loader=FileSystemLoader(scratch_dir), **environment_kwargs)
  elif template_dir:
    custom_template_dir = os.path.abspath(template_dir)
    env = Environment(loader=FileSystemLoader(os.path.abspath(template_dir)), **environment_kwargs)
  else:
    env = Environment(loader=FileSystemLoader(templates_dir_path()))

  required_templates = [
    'Dockerfile.j2',
    '.dockerignore.j2',
    'hokusai/build.yml.j2',
    'hokusai/development.yml.j2',
    'hokusai/test.yml.j2',
    'hokusai/staging.yml.j2',
    'hokusai/production.yml.j2'
  ]

  template_context = {
    "project_name": config.project_name,
    "project_repo": ecr.project_repo
  }

  for s in template_vars:
    if '=' not in s:
      raise HokusaiError("Error: template variables must be of the form 'key=value'")
    split = s.split('=', 1)
    template_context[split[0]] = split[1]

  try:
    for template in required_templates:
      if custom_template_dir and not os.path.isfile(os.path.join(custom_template_dir, template)):
        raise HokusaiError("Could not find required template file %s" % template)
      with open(os.path.join(CWD, template.rstrip('.j2')), 'w') as f:
        f.write(env.get_template(template).render(**template_context))
      print_green("Created %s" % template.rstrip('.j2'))

    if custom_template_dir:
      for root, _, files in os.walk(custom_template_dir):
        subpath = os.path.relpath(root, custom_template_dir)
        if subpath != '.':
          mkpath(os.path.join(CWD, subpath))
        for file in files:
          if subpath != '.':
            file_path = os.path.join(subpath, file)
          else:
            file_path = file
          if file_path in required_templates:
            continue
          if file_path.endswith('.j2'):
            with open(os.path.join(CWD, file_path.rstrip('.j2')), 'w') as f:
              f.write(env.get_template(file_path).render(**template_context))
          else:
            copyfile(os.path.join(custom_template_dir, file_path), os.path.join(CWD, file_path))
          print_green("Created %s" % file_path.rstrip('.j2'))
  finally:
    if scratch_dir:
      rmtree(scratch_dir)



import os
import yaml

from hokusai.lib.common import (
  local_to_local,
  print_red,
  unlink_file_if_not_debug,
  write_temp_file,
  yaml_content_with_header
)
from hokusai.lib.config import HOKUSAI_TMP_DIR
from hokusai.lib.config_loader import ConfigLoader
from hokusai.lib.exceptions import HokusaiError


user_home = os.environ.get('HOME', '/')
local_global_config_file_name = '.hokusai.yml'
local_global_config = os.path.join(
  user_home,
  local_global_config_file_name
)
source_global_config = os.environ.get(
  'HOKUSAI_GLOBAL_CONFIG',
  local_global_config
)

class HokusaiGlobalConfig:
  ''' manage Hokusai global config '''
  def __init__(self):
    self._config = ConfigLoader(source_global_config).load()
    self.validate_config()

  def merge(self, **kwargs):
    ''' merge key/value pairs into config, skipping None '''
    for k,v in kwargs.items():
      if v is not None:
        # use dash in _config keys
        self._config[k.replace('_', '-')] = v

  def save(self):
    ''' save config to local config file '''
    try:
      config_string = yaml.safe_dump(
        self._config,
        default_flow_style=False
      )
      tmp_path = write_temp_file(
        yaml_content_with_header(config_string),
        HOKUSAI_TMP_DIR
      )
      local_to_local(
        tmp_path,
        user_home,
        local_global_config_file_name,
        create_target_dir=False
      )
      unlink_file_if_not_debug(tmp_path)
    except:
      print_red(
        f'Error: Not able to write Hokusai config to {local_global_config}'
      )
      raise

  def validate_config(self):
    ''' sanity check config '''
    required_vars = [
      'kubectl-version',
      'kubeconfig-dir',
      'kubeconfig-source-uri',
      'kubectl-dir'
    ]
    for var in required_vars:
      if not var in self._config:
        raise HokusaiError(
          f'{var} is missing in Hokusai global config'
        )

  @property
  def kubeconfig_dir(self):
    return self._config['kubeconfig-dir']

  @property
  def kubeconfig_source_uri(self):
    return self._config['kubeconfig-source-uri']

  @property
  def kubectl_dir(self):
    return self._config['kubectl-dir']

  @property
  def kubectl_version(self):
    return self._config['kubectl-version']



from hokusai.lib.config import config
from hokusai.lib.common import shout, shout_concurrent
from hokusai.services.kubectl import Kubectl
from hokusai.lib.exceptions import HokusaiError

def logs(context, timestamps, follow, tail, previous, labels, namespace=None):
  kctl = Kubectl(context, namespace=namespace)

  opts = ''
  if timestamps:
    opts += ' --timestamps'
  if previous:
    opts += ' --previous'
  if follow or config.follow_logs:
    opts += ' --follow'
  if tail or config.tail_logs:
    num_tail = tail if tail else config.tail_logs
    opts += " --tail=%s" % num_tail

  selectors = ["app=%s" % config.project_name, "layer=application"]
  for l in labels:
    if '=' not in l:
      raise HokusaiError("Error: label selectors of the form 'key=value'")
    selectors.append(l)

  pods = kctl.get_objects('pod', selector=(',').join(selectors))
  pods = [pod for pod in pods if pod['status']['phase'] == 'Running']
  containers = []

  for pod in pods:
    for container in pod['spec']['containers']:
      containers.append({'pod': pod['metadata']['name'], 'name': container['name']})

  commands = [kctl.command("logs %s %s%s" % (container['pod'], container['name'], opts)) for container in containers]
  shout_concurrent(commands, print_output=True)
